local Animation = {}
ShadowUF:RegisterModule(Animation, "smoothBars", "Smooth bar changes")

local animations = {}
local activeAnimations = {}

local f = CreateFrame("Frame")
local ipairs = _G.ipairs
local GetTime = _G.GetTime
local tinsert, tremove = _G.tinsert, _G.tremove

--[[
  Adapted from the jQuery easing library under its MIT license
  
  c = change [total amount]
  t = time
  b = beginning value
  d = total animation time
]]--
local animationStyleOptions = {
	LINEAR = "Linear",
	EASE_IN = "Ease (in)",
	EASE_OUT = "Ease (out)",
	EASE_IN_OUT = "Ease (in & out)",
	EXPONENTIAL_IN = "Exponential (in)",
	EXPONENTIAL_OUT = "Exponential (out)"	
}

local animationStyles = {
	LINEAR = function(t, b, c, d)
		return c*t/d + b
	end,
	EASE_IN = function(t, b, c, d)
		t = t / d
		return c * t * t + b
	end,
	EASE_OUT = function(t, b, c, d)
		return -c*t*t/(d*d) + 2*c*t/d + b
	end,
	EASE_IN_OUT = function(t, b, c, d)
		if (t < d/2) then
			return 2*c*t*t/(d*d) + b
		end
		local ts = t - d/2
		return -2*c*ts*ts/(d*d) + 2*c*ts/d + c/2 + b		
	end,
	EXPONENTIAL_IN = function(t, b, c, d)
		local flip = 1
		if c < 0 then
			flip = -flip
			c = -c
		end
		return flip * (math.exp(math.log(c)/d * t)) + b	
	end,
	EXPONENTIAL_OUT = function(t, b, c, d)
		local flip = 1
		if c < 0 then
			flip = -flip
			c = -c
		end
		return flip * (-math.exp(-math.log(c)/d * (t-d)) + c + 1) + b
	end
}

local function animateAll(f, t)
	local ct = GetTime()
	for index, bar in ipairs(animations) do
		local delta = ct - bar.animateStartTime
		local total = bar.animationSpeed
		if delta >= total then
			bar:OrigSetValue(bar.animateEndValue)
		else
			local amt = bar.animationFunc(delta, bar.animateStartValue, bar.animateDelta, total)
			bar:OrigSetValue(amt)
		end
	end
end

local function SetValue(bar, value)
	-- Don't currently have a way to set this.
	-- This intent is to avoid animation on target switch during initial value population.
	if bar.parent.ignoreAnimation then
		bar.parent.ignoreAnimation = nil
		bar:OrigSetValue(value)
		return
	end
	
	-- Ignore animation for updates that are <1% of the total
	local initial = bar:GetValue()
	local mn, mx = bar:GetMinMaxValues()
	if math.abs(value - initial) / mx <= 0.01 then
		bar:OrigSetValue(value)
		return
	end		
	
	bar.animateStartTime = GetTime()
	bar.animateStartValue = initial
	
	bar.animateEndValue = value
	bar.animateDelta = bar.animateEndValue - bar.animateStartValue
	if not activeAnimations[bar] then
		activeAnimations[bar] = true
		tinsert(animations, bar)
		f:SetScript("OnUpdate", animateAll)
	end
end

local function registerForAnimation(bar, speed, animationFunc)
	if bar.SetValue ~= SetValue then
		bar.OrigSetValue = bar.SetValue
		bar.SetValue = SetValue
	end
	bar.animationSpeed = speed
	bar.animationFunc = animationFunc or Animation.AnimationFunctions.LINEAR
end

function unregisterForAnimation(bar)
	bar.SetValue = bar.OrigSetValue
	bar.OrigSetValue = nil
	bar.animationSpeed = nil
	bar.animationFunc = nil
	if activeAnimations[bar] then
		for k, v in ipairs(animations) do
			if v == bar then
				tremove(animations, k)
				break
			end
		end
		activeAnimations[bar] = nil
	end
	
	if bar.animateEndValue and bar:GetValue() ~= bar.animateEndValue then
		bar:SetValue(bar.animateEndValue)
	end
end

function Animation:OnConfigurationLoad()
	ShadowUF.Config.unitTable.args.bars.args.smoothBars = {
		type = "group",
		inline = true,
		name = "Smooth bar fading",
		args = {
			speed = {
				name = "Speed",
				type = "range",
				min = 0.05,
				max = 1.0,
				step = 0.01,
				bigStep = 0.05,
				arg = "smoothBars.speed"
			},
			style = {
				type = "select",
				name = "Animation style",
				values = animationStyleOptions,
				arg = "smoothBars.style"
			},
			enabled = {
				name = "Enable SmoothBars",
				type = "toggle",
				arg = "smoothBars.enabled",
				order = 1
			}
		}
	}
end

function Animation:OnDefaultsSet()
	for _, unit in pairs(ShadowUF.units) do
		ShadowUF.defaults.profile.units[unit].smoothBars = {
			enabled = true,
			speed = 0.25,
			style = "LINEAR"
		}
	end	
end

do	
	local bars = {"healthBar", "powerBar", "xpBar"}
	function Animation:OnLayoutApplied(frame)
		if not frame.visibility.smoothBars then return end
		for _, bar in ipairs(bars) do
			if frame[bar] then
				local opts = ShadowUF.db.profile.units[frame.unitType].smoothBars
				if opts.enabled then
					registerForAnimation(frame[bar], opts.speed, animationStyles[opts.style])
				else
					unregisterForAnimation(frame[bar])
				end
			end
		end
	end
	
	function Animation:OnEnable(frame)
		self:OnLayoutApplied(frame)
	end

	function Animation:OnDisable(frame)	
		for _, bar in ipairs(bars) do
			if frame[bar] then
				unregisterForAnimation(frame[bar])
			end
		end
	end
end
